﻿Imports System
Imports System.Collections.Generic
' Extracting frequent item-sets for Association Rule Learning.

' Demo of extracting frequent item-sets from a set of transactions.
' Extracting frequent item-sets is useful by itself, but also is
' the first phase of 'assocation rules learning' (not covered in this demo).
' Demo uses a variation of the 'Apriori' algorithm.

Namespace FreqItemSets
  Friend Class FreqItemSetProgram
	Shared Sub Main(ByVal args() As String)
	  Try
		Console.WriteLine(VBLf & "Begin frequent item-set extraction demo" & VBLf)

		Dim rawItems() As String = { "apples", "bread ", "celery", "donuts", "eggs  ", "flour ", "grapes", "honey ", "icing ", "jelly ", "kale  ", "lettuce" }

		Dim N As Integer = rawItems.Length ' total number of items to deal with ( [0..11] )

		Dim rawTransactions(9)() As String
		rawTransactions(0) = New String() { "apples", "bread ", "celery", "flour " } ' 0 1 2 5
		rawTransactions(1) = New String() { "bread ", "eggs  ", "flour " } ' 1 4 5
		rawTransactions(2) = New String() { "apples", "bread ", "donuts", "eggs  " } ' 0 1 3 4
		rawTransactions(3) = New String() { "celery", "donuts", "flour ", "grapes" } ' 2 3 5 6
		rawTransactions(4) = New String() { "donuts", "eggs  " } ' 3 4
		rawTransactions(5) = New String() { "donuts", "eggs  ", "jelly " } ' 3 4 9
		rawTransactions(6) = New String() { "apples", "bread ", "donuts", "icing " } ' 0 1 3 8
		rawTransactions(7) = New String() { "bread ", "grapes", "honey " } ' 1 6 7
		rawTransactions(8) = New String() { "apples", "bread ", "celery", "flour ", "kale  " } ' 0 1 2 5 10
		rawTransactions(9) = New String() { "apples", "bread ", "celery", "flour " } ' 0 1 2 5

		Console.WriteLine("Raw transactions are:")
		Console.WriteLine("-----------------------------------------------")
		For i As Integer = 0 To rawTransactions.Length - 1
		  Console.Write("[" & i & "] : ")
		  For j As Integer = 0 To rawTransactions(i).Length - 1
			Console.Write(rawTransactions(i)(j) & "   ")
		  Next j
		  Console.WriteLine("")
		Next i

		' could do this programmatically
		Dim transactions As New List(Of Integer())()
		transactions.Add(New Integer() { 0, 1, 2, 5 })
		transactions.Add(New Integer() { 1, 4, 5 })
		transactions.Add(New Integer() { 0, 1, 3, 4 })
		transactions.Add(New Integer() { 2, 3, 5, 6 })
		transactions.Add(New Integer() { 3, 4 })
		transactions.Add(New Integer() { 3, 4, 9 })
		transactions.Add(New Integer() { 0, 1, 3, 8 })
		transactions.Add(New Integer() { 1, 6, 7 })
		transactions.Add(New Integer() { 0, 1, 2, 5, 10 })
		transactions.Add(New Integer() { 0, 1, 2, 5 })

		Console.WriteLine(vbLf & vbLf & "Encoded transactions are:")
		Console.WriteLine("-----------------------------------------------")
		For i As Integer = 0 To transactions.Count - 1
		  Console.Write("[" & i & "] : ")
		  For j As Integer = 0 To transactions(i).Length - 1
			Console.Write(transactions(i)(j).ToString() & " ")
		  Next j
		  Console.WriteLine("")
		Next i
		Console.WriteLine("")


		Dim minSupportPct As Double = 0.30 ' minimum pct of transactions for an item-set to be 'frequent'
		Dim minItemSetLength As Integer = 2
		Dim maxItemSetLength As Integer = 4
		Console.WriteLine(vbLf & "Setting minimum frequent support percent = " & minSupportPct.ToString("F2"))
		Console.WriteLine("Setting minimum frequent item-set length = " & minItemSetLength)
		Console.WriteLine("Setting maximum frequent item-set length = " & maxItemSetLength)
		Console.WriteLine("Using Apriori algorithm to construct frequent item-sets")

		' everything happens here
		Dim frequentItemSets As List(Of ItemSet) = GetFrequentItemSets(N, transactions, minSupportPct, minItemSetLength, maxItemSetLength)

		Console.WriteLine(VBLf & "Frequent item-sets in numeric form are:")
		Console.WriteLine("")
		For i As Integer = 0 To frequentItemSets.Count - 1
		  Console.WriteLine(frequentItemSets(i).ToString())
		Next i

		Console.WriteLine(VBLf & "Frequent item-sets in string form are:")
		Console.WriteLine("")
		For i As Integer = 0 To frequentItemSets.Count - 1
		  For j As Integer = 0 To frequentItemSets(i).data.Length - 1
			Dim v As Integer = frequentItemSets(i).data(j)
			Console.Write(rawItems(v) & "   ")
		  Next j
		  Console.WriteLine("")
		Next i

		Console.WriteLine(VBLf & "End frequent item-set extraction demo" & VBLf)
		Console.ReadLine()
	  Catch ex As Exception
		Console.WriteLine(ex.Message)
		Console.ReadLine()
	  End Try
	End Sub ' Main

	Private Shared Function GetFrequentItemSets(ByVal N As Integer, ByVal transactions As List(Of Integer()), ByVal minSupportPct As Double, ByVal minItemSetLength As Integer, ByVal maxItemSetLength As Integer) As List(Of ItemSet)
	  ' create a List of frequent ItemSet objects that are in transactions
	  ' frequent means occurs in minSupportPct percent of transactions
	  ' N is total number of items
	  ' uses a variation of the Apriori algorithm

	  Dim minSupportCount As Integer = CInt(Fix(transactions.Count * minSupportPct))

	  Dim frequentDict As New Dictionary(Of Integer, Boolean)() ' key = int representation of an ItemSet, val = is in List of frequent ItemSet objects
	  Dim frequentList As New List(Of ItemSet)() ' item set objects that meet minimum count (in transactions) requirement
	  Dim validItems As New List(Of Integer)() ' inidividual items/values at any given point in time to be used to construct new ItemSet (which may or may not meet threshhold count)

	  ' get counts of all individual items
	  Dim counts(N - 1) As Integer ' index is the item/value, cell content is the count
	  For i As Integer = 0 To transactions.Count - 1
		For j As Integer = 0 To transactions(i).Length - 1
		  Dim v As Integer = transactions(i)(j)
		  counts(v) += 1
		Next j
	  Next i
	  ' for those items that meet support threshold, add to valid list, frequent list, frequent dict
	  For i As Integer = 0 To counts.Length - 1
		If counts(i) >= minSupportCount Then ' frequent item
		  validItems.Add(i) ' i is the item/value
		  Dim d(0) As Integer ' the ItemSet ctor wants an array
		  d(0) = i
		  Dim ci As New ItemSet(N, d, 1) ' an ItemSet with size 1, ct 1
		  frequentList.Add(ci) ' it's frequent
		  frequentDict.Add(ci.hashValue, True)
		End If ' else skip this item
	  Next i

	  Dim done As Boolean = False ' done if no new frequent item-sets found
	  Dim k As Integer = 2
	  Do While k <= maxItemSetLength AndAlso done = False ' construct all size k = 2, 3, 4,.. frequent item-sets
		done = True ' assume no new item-sets will be created
		Dim numFreq As Integer = frequentList.Count ' List size modified so store first

		For i As Integer = 0 To numFreq - 1 ' use existing frequent item-sets to create new freq item-sets with size+1
		  If frequentList(i).k <> k - 1 Then ' only use those ItemSet objects with size 1 less than new ones being created
			  Continue For
		  End If

		  Dim j As Integer = 0
		  Do While j < validItems.Count
			Dim newData(k - 1) As Integer ' data for a new item-set

			For p As Integer = 0 To k - 2
			  newData(p) = frequentList(i).data(p) ' old data in
			Next p

			If validItems(j) <= newData(k - 2) Then ' because item-values are in order we can skip sometimes
				j += 1
				Continue Do
			End If

			newData(k - 1) = validItems(j) ' new item-value
			Dim ci As New ItemSet(N, newData, -1) ' ct to be determined

			If frequentDict.ContainsKey(ci.hashValue) = True Then ' this new ItemSet has already been added
			  j += 1
			  Continue Do
			End If
			Dim ct As Integer = CountTimesInTransactions(ci, transactions) ' how many times is the new ItemSet in the transactuions?
			If ct >= minSupportCount Then ' we have a winner!
			  ci.ct = ct ' now we know the ct
			  frequentList.Add(ci)
			  frequentDict.Add(ci.hashValue, True)
			  done = False ' a new item-set was created, so we're not done
			End If
			  j += 1
		  Loop ' j
		Next i ' i

		' update valid items -- quite subtle
		validItems.Clear()
		Dim validDict As New Dictionary(Of Integer, Boolean)() ' track new list of valid items
		For idx As Integer = 0 To frequentList.Count - 1
		  If frequentList(idx).k <> k Then ' only looking at the just-created item-sets
			  Continue For
		  End If
		  For j As Integer = 0 To frequentList(idx).data.Length - 1
			Dim v As Integer = frequentList(idx).data(j) ' item
			If validDict.ContainsKey(v) = False Then
			  validItems.Add(v)
			  validDict.Add(v, True)
			End If
		  Next j
		Next idx
		validItems.Sort() ' keep valid item-values ordered so item-sets will always be ordered
		  k += 1
	  Loop ' next k

	  ' transfer to return result, filtering by minItemSetCount
	  Dim result As New List(Of ItemSet)()
	  For i As Integer = 0 To frequentList.Count - 1
		If frequentList(i).k >= minItemSetLength Then
		  result.Add(New ItemSet(frequentList(i).N, frequentList(i).data, frequentList(i).ct))
		End If
	  Next i

	  Return result
	End Function

	Private Shared Function CountTimesInTransactions(ByVal itemSet As ItemSet, ByVal transactions As List(Of Integer())) As Integer
	  ' number of times itemSet occurs in transactions
	  Dim ct As Integer = 0
	  For i As Integer = 0 To transactions.Count - 1
		If itemSet.IsSubsetOf(transactions(i)) = True Then
		  ct += 1
		End If
	  Next i
	  Return ct
	End Function
  End Class ' program

  Public Class ItemSet
	Public N As Integer ' data items are [0..N-1]
	Public k As Integer ' number of items
	Public data() As Integer ' ex: [0 2 5]
	Public hashValue As Integer ' "0 2 5" -> 520 (for hashing)
	Public ct As Integer ' num times this occurs in transactions

	Public Sub New(ByVal N As Integer, ByVal items() As Integer, ByVal ct As Integer)
	  Me.N = N
	  Me.k = items.Length
	  Me.data = New Integer(Me.k - 1){}
	  Array.Copy(items, Me.data, items.Length)
	  Me.hashValue = ComputeHashValue(items)
	  Me.ct = ct
	End Sub

	Private Shared Function ComputeHashValue(ByVal data() As Integer) As Integer
	  Dim value As Integer = 0
	  Dim multiplier As Integer = 1
	  For i As Integer = 0 To data.Length - 1 ' actually working backward
		value = value + (data(i) * multiplier)
		multiplier = multiplier * 10
	  Next i
	  Return value
	End Function

	Public Overrides Function ToString() As String
	  Dim s As String = "{ "
	  For i As Integer = 0 To data.Length - 1
		s &= data(i) & " "
	  Next i
	  Return s & "}" & "   ct = " & ct

	End Function

	Public Function IsSubsetOf(ByVal trans() As Integer) As Boolean
	  ' 'trans' is an ordered transaction like [0 1 4 5 8]
	  Dim foundIdx As Integer = -1
	  For j As Integer = 0 To Me.data.Length - 1
		foundIdx = IndexOf(trans, Me.data(j), foundIdx + 1)
		If foundIdx = -1 Then
			Return False
		End If
	  Next j
	  Return True
	End Function
	Private Shared Function IndexOf(ByVal array() As Integer, ByVal item As Integer, ByVal startIdx As Integer) As Integer
	  For i As Integer = startIdx To array.Length - 1
		If i > item Then ' i is past where the target could possibly be
			Return -1
		End If
		If array(i) = item Then
			Return i
		End If
	  Next i
	  Return -1
	End Function


  End Class ' ItemSet

End Namespace ' ns
